#!/usr/bin/env python3

'''Common classes for the project.'''

from typing import Any, Iterator, ItemsView, KeysView, Mapping, ValuesView, SupportsBytes, Union

from .header_values import HeaderValue
from .incremental_parser import parse_header_value


class SpamcBody:
    '''Provides a descriptor for a bytes-like object.'''

    def __get__(self, instance, owner) -> bytes:
        return instance._body

    def __set__(self, instance, value: Union[bytes, SupportsBytes]) -> None:
        instance._body = bytes(value)


class SpamcHeaders:
    '''Provides a dictionary-like interface for headers.'''

    def __init__(self, *, headers: Mapping[str, Union[HeaderValue, str, Any]] = None) -> None:
        self._headers = {}
        if headers:
            for key, value in headers.items():
                self[key] = value

    def __str__(self):
        return '<{} object at {}, keys: {}>'.format(
            '.'.join([self.__class__.__module__, self.__class__.__qualname__]),
            id(self),
            ', '.join(self._headers.keys())
        )

    def __bytes__(self) -> bytes:
        return b''.join(
            [b'%b: %b\r\n' % (name.encode('ascii'), bytes(value)) for name, value in self._headers.items()]
        )

    def __getitem__(self, key: str) -> HeaderValue:
        return self._headers[key]

    def __setitem__(self, key: str, value: Union[HeaderValue, str]) -> None:
        if isinstance(value, HeaderValue):
            self._headers[key] = value
        else:
            self._headers[key] = parse_header_value(key, value)

    def __iter__(self) -> Iterator[str]:
        return iter(self._headers)

    def __len__(self) -> int:
        return len(self._headers)

    def keys(self) -> KeysView[str]:
        return self._headers.keys()

    def items(self) -> ItemsView[str, HeaderValue]:
        return self._headers.items()

    def values(self) -> ValuesView[HeaderValue]:
        return self._headers.values()



#!/usr/bin/env python3

'''Contains the Client class that is used to interact with SPAMD.'''

import asyncio
import logging
from pathlib import Path
import ssl
from typing import SupportsBytes, Union

import certifi

from .exceptions import BadResponse, ResponseException
from .incremental_parser import ResponseParser, ParseError
from .options import ActionOption, MessageClassOption
from .requests import Request
from .responses import Response


class Client:
    '''Client object for interacting with SPAMD.'''

    def __init__(self,
                 socket_path: str = None,
                 host: str = 'localhost',
                 port: int = 783,
                 user: str = None,
                 compress: bool = False,
                 verify: Union[bool, str, Path] = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        '''Client constructor.

        :param socket_path: The path to the Unix socket for the SPAMD service.
        :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
        :param port: Port number for the SPAMD service, defaults to 783.
        :param user: Name of the user that SPAMD will run the checks under.
        :param compress: If true, the request body will be compressed.
        :param verify: Use SSL for the connection.  If True, will use root certificates.
            If False, will not verify the certificate.  If a string to a path or a Path
            object, the connection will use the certificates found there.
        :param loop: The asyncio event loop.

        :raises ValueError: Raised if the constructor can't tell if it's using a TCP or a Unix domain socket connection.
        '''

        if socket_path:
            from aiospamc.connections.unix_connection import UnixConnectionManager
            self.connection = UnixConnectionManager(socket_path, loop=loop)
        elif host and port:
            from aiospamc.connections.tcp_connection import TcpConnectionManager
            if verify is not None:
                self.connection = TcpConnectionManager(host, port, self.new_ssl_context(verify), loop=loop)
            else:
                self.connection = TcpConnectionManager(host, port, loop=loop)
        else:
            raise ValueError('Either "host" and "port" or "socket_path" must be specified.')

        self._host = host
        self._port = port
        self._socket_path = socket_path
        self.user = user
        self.compress = compress

        self.logger = logging.getLogger(__name__)
        self.logger.debug('Created instance of %r', self)

    def __repr__(self) -> str:
        client_fmt = ('{}(socket_path={}, '
                      'host={}, '
                      'port={}, '
                      'user={}, '
                      'compress={})')
        return client_fmt.format(self.__class__.__name__,
                                 repr(self._socket_path),
                                 repr(self._host),
                                 repr(self._port),
                                 repr(self.user),
                                 repr(self.compress))

    @staticmethod
    def new_ssl_context(value: Union[bool, str, Path]) -> ssl.SSLContext:
        '''Creates an SSL context based on the supplied parameter.

        :param value: Use SSL for the connection.  If True, will use root certificates.
            If False, will not verify the certificate.  If a string to a path or a Path
            object, the connection will use the certificates found there.
        '''

        if value is True:
            return ssl.create_default_context(cafile=certifi.where())
        elif value is False:
            context = ssl.create_default_context(cafile=certifi.where())
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            return context

        cert_path = Path(value).absolute()
        if cert_path.is_dir():
            return ssl.create_default_context(capath=str(cert_path))
        elif cert_path.is_file():
            return ssl.create_default_context(cafile=str(cert_path))
        else:
            raise FileNotFoundError('Certificate path does not exist at {}'.format(value))

    async def send(self, request: Request) -> Response:
        '''Sends a request to the SPAMD service.

        If the SPAMD service gives a temporary failure response, then its retried.

        :param request: Request object to send.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        if self.compress:
            request.headers['Compress'] = 'zlib'
        if self.user:
            request.headers['User'] = self.user

        self.logger.debug('Sending request (%s)', id(request))
        async with self.connection.new_connection() as connection:
            await connection.send(bytes(request))
            self.logger.debug('Request (%s) successfully sent', id(request))
            data = await connection.receive()

        try:
            try:
                parser = ResponseParser()
                parsed_response = parser.parse(data)
                response = Response(**parsed_response)
            except ParseError:
                raise BadResponse
            response.raise_for_status()
        except ResponseException as error:
            self.logger.exception('Exception for request (%s)when composing response: %s',
                                  id(request),
                                  error)
            raise

        self.logger.debug('Received response (%s) for request (%s)',
                          id(response),
                          id(request))
        return response

    async def check(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Request the SPAMD service to check a message.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('CHECK', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def headers(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Request the SPAMD service to check a message with a HEADERS request.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.  The body contains the modified
            message headers, but not the content of the message.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('HEADERS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def ping(self) -> Response:
        '''Sends a ping request to the SPAMD service and will receive a
        response if the service is alive.

        :return: A response with "PONG".

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('PING')
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def process(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Process the message and return a modified copy of the message.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.  The body contains a modified
            copy of the message.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('PROCESS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def report(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Check if message is spam and return report.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.  The body contains a report.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('REPORT', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def report_if_spam(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Check if a message is spam and return a report if the message is spam.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.  The body contains a report if
            the message is considered spam.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('REPORT_IFSPAM', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def symbols(self, message: Union[bytes, SupportsBytes]) -> Response:
        '''Check if the message is spam and return a list of symbols that were hit.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        :return:
            A successful response with a "Spam" header showing if the message is
            considered spam as well as the score.  The body contains a
            comma-separated list of the symbols that were hit.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request('SYMBOLS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def tell(self,
                   message: Union[bytes, SupportsBytes],
                   message_class: Union[str, MessageClassOption],
                   remove_action: Union[str, ActionOption] = None,
                   set_action: Union[str, ActionOption] = None,
                   ):
        '''Instruct the SPAMD service to to mark the message.

        :param message:
            A byte string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.
        :param message_class: How to classify the message, either "ham" or "spam".
        :param remove_action: Remove message class for message in database.
        :param set_action:
            Set message class for message in database.  Either `ham` or `spam`.

        :return:
            A successful response with "DidSet" and/or "DidRemove" headers along with the
            actions that were taken.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises TimeoutException: Timeout during connection.
        '''

        request = Request(verb='TELL', body=message)

        request.headers['Message-class'] = message_class

        if remove_action:
            request.headers['Remove'] = remove_action
        if set_action:
            request.headers['Set'] = set_action
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response



#!/usr/bin/env python3

'''Module for the parsing functions and objects.'''

from enum import Enum
import re
from typing import Any, Callable, Mapping, Tuple, Union, Dict

from .header_values import CompressValue, ContentLengthValue, GenericHeaderValue, HeaderValue, MessageClassValue, \
    SetOrRemoveValue, SpamValue, UserValue
from .options import ActionOption, MessageClassOption


class ParseError(Exception):
    '''Error occurred while parsing.'''

    def __init__(self, message=None) -> None:
        '''Construct parsing exception with optional message.

        :param message: User friendly message.
        '''

        self.message = message


class NotEnoughDataError(ParseError):
    '''Expected more data than what the protocol content specified.'''

    pass


class TooMuchDataError(ParseError):
    '''Too much data was received than what the protocol content specified.'''
    pass


class States(Enum):
    '''States for the parser state machine.'''

    Status = 1
    Header = 2
    Body = 3
    Done = 4


class Parser:
    '''The parser state machine.

    :ivar result: Storage location for parsing results.'''

    def __init__(self,
                 delimiter: bytes,
                 status_parser: Callable[[bytes], Mapping[str, str]],
                 header_parser: Callable[[bytes], Tuple[str, Any]],
                 body_parser: Callable[[bytes, int], bytes],
                 start: States = States.Status) -> None:
        '''Parser constructor.

        :param delimiter: Byte string to split the different sections of the message.
        :param status_parser: Callable to parse the status line of the message.
        :param header_parser: Callable to parse each header line of the message.
        :param body_parser: Callable to parse the body of the message.
        :param start: The state to start the parser on.  Allowed for easier testing.
        '''

        self.delimiter = delimiter
        self.status_parser = status_parser
        self.header_parser = header_parser
        self.body_parser = body_parser
        self.result = {'headers': {}, 'body': b''}

        self._state = start
        self.buffer = b''

        self._state_table = {
            States.Status: self.status,
            States.Header: self.header,
            States.Body: self.body
        }

    @property
    def state(self) -> States:
        '''The current state of the parser.

        :return: The :class:`States` instance.
        '''

        return self._state

    def parse(self, stream: bytes) -> Mapping[str, Any]:
        '''Entry method to parse a message.

        :param stream: Byte string to parse.

        :return: Returns the parser results dictionary stored in the class attribute :attr:`result`.

        :raises NotEnoughDataError: Raised when not enough data is sent to be parsed.
        :raises TooMuchDataError: Raised when too much data is sent to be parsed.
        :raises ParseError: Raised when a general parse error is found.
        '''

        self.buffer += stream

        while self._state != States.Done:
            if self._state == States.Status:
                self.status()
            if self._state == States.Header:
                self.header()
            if self._state == States.Body:
                self.body()

        return self.result

    def status(self) -> None:
        '''Splits the message at the delimiter and sends the first part of the message to the `status_line` callable to
        be parsed.  If successful then the results are stored in the :attr:`result` class attribute and the state
        transitions to :class:`States.Header`.

        :raises NotEnoughDataError: When there is no delimiter the message is incomplete.
        :raises ParseError: When the `status_line` callable experiences an error.
        '''

        status_line, delimiter, leftover = self.buffer.partition(self.delimiter)

        if status_line and delimiter:
            self.buffer = leftover
            self.result = {**self.result, **self.status_parser(status_line)}
            self._state = States.Header
        else:
            raise NotEnoughDataError

    def header(self) -> None:
        '''Splits the message at the delimiter and sends the line to the `header_parser`.

        When splitting the action will be determined depending what is matched:

        +-------------+-----------+-----------+----------------------------------------------------------------------+
        | Header line | Delimiter | Leftover  | Action                                                               |
        +=============+===========+===========+======================================================================+
        |     No      |    Yes    | Delimiter | Headers done, empty body. Clear buffer and transition to             |
        |             |           |           | :class:`States.Body`.                                                |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     No      |    Yes    |    N/A    | Headers done.  Transition to :class:`States.Body`.                   |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     Yes     |    Yes    |    N/A    | Parse header.  Record in :attr:`result` class attribute.             |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     No      |    No     |    No     | Message was a status line only.  Transition to :class:`States.Body`. |
        +-------------+-----------+-----------+----------------------------------------------------------------------+

        :raises ParseError: None of the previous conditions are matched.
        '''

        header_line, delimiter, leftover = self.buffer.partition(self.delimiter)

        if not header_line and delimiter and leftover == self.delimiter:
            self.buffer = b''
            self._state = States.Body
        elif not header_line and delimiter:
            self.buffer = leftover
            self._state = States.Body
        elif header_line and delimiter:
            self.buffer = leftover
            key, value = self.header_parser(header_line)
            self.result['headers'][key] = value
        elif not header_line and not delimiter and not leftover:
            self._state = States.Body
        else:
            raise ParseError('Header section not in recognizable format')

    def body(self) -> None:
        '''Uses the length defined in the `Content-length` header (defaulted to 0) to determine how many bytes the body
        contains.

        :raises TooMuchDataError:
            When there are too many bytes in the buffer compared to the `Content-length` header value.  Transitions the
            state to :class:`States.Done`.
        '''

        content_length = self.result['headers'].get('Content-length', ContentLengthValue(length=0)).length
        try:
            self.result['body'] += self.body_parser(self.buffer, content_length)
            self._state = States.Done
        except TooMuchDataError:
            self._state = States.Done
            raise


def parse_request_status(stream: bytes) -> Dict[str, str]:
    '''Parses the status line from a request.

    :param stream: The byte stream to parse.

    :return: A dictionary with the keys `verb`, `protocol` and `version`.

    :raises ParseError:
        When the status line is in an invalid format, not a valid verb, or doesn't have the correct protocol.
    '''

    try:
        verb, protocol_version = stream.decode('ascii').split(' ')
        protocol, version = protocol_version.split('/')
    except ValueError:
        raise ParseError('Could not parse request status line, not in recognizable format')

    if verb not in ['CHECK', 'HEADERS', 'PING', 'PROCESS', 'REPORT_IFSPAM', 'REPORT', 'SKIP', 'SYMBOLS', 'TELL']:
        raise ParseError('Not a valid verb')

    if protocol != 'SPAMC':
        raise ParseError('Protocol name does not match')

    return {'verb': verb, 'protocol': protocol, 'version': version}


def parse_response_status(stream: bytes) -> Dict[str, str]:
    '''Parse the status line for a response.

    :param stream: The byte stream to parse.

    :return: A dictionary with the keys `protocol`, `version`, `status_code`, and `message`.

    :raises ParseError:
        When the status line is in an invalid format, status code is not an integer, or doesn't have the correct
        protocol.
    '''

    try:
        protocol_version, status_code, message = list(filter(None, stream.decode('ascii').split(' ')))
        protocol, version = protocol_version.split('/')
    except ValueError:
        raise ParseError('Could not parse response status line, not in recognizable format')

    if protocol != 'SPAMD':
        raise ParseError('Protocol name does not match')

    try:
        status_code = int(status_code)
    except ValueError:
        raise ParseError('Protocol status code is not an integer')

    return {
        'protocol': protocol,
        'version': version,
        'status_code': status_code,
        'message': message
    }


def parse_message_class_value(stream: Union[str, MessageClassOption]) -> MessageClassValue:
    '''Parses the `Message-class` header value.

    :param stream: String or `MessageClassOption` instance.

    :return: A `MessageClassValue` instance representing the value.

    :raises ParseError: When the value doesn't match either `ham` or `spam`.
    '''

    try:
        stream = stream.name
    except AttributeError:
        pass

    stream = stream.strip()

    try:
        return MessageClassValue(value=getattr(MessageClassOption, stream))
    except AttributeError:
        raise ParseError('Unable to parse Message-class header value')


def parse_content_length_value(stream: Union[str, int]) -> ContentLengthValue:
    '''Parses the `Content-length` header value.

    :param stream: String or integer value of the header.

    :return: A `ContentLengthValue` instance.

    :raises ParseError: When the value cannot be cast to an integer.
    '''

    try:
        value = int(stream)
    except ValueError:
        raise ParseError('Unable to parse Content-length value, must be integer')

    return ContentLengthValue(length=value)


def parse_compress_value(stream: str) -> CompressValue:
    '''Parses a value for the `Compress` header.

    :param stream: String to parse.

    :return: A `CompressValue` instance.
    '''

    return CompressValue(algorithm=stream.strip())


def parse_set_remove_value(stream: Union[ActionOption, str]) -> SetOrRemoveValue:
    '''Parse a value for the `DidRemove`, `DidSet`, `Remove`, and `Set` headers.

    :param stream: String to parse or an instance of `ActionOption`.

    :return: A `SetOrRemoveValue` instance.
    '''

    if isinstance(stream, ActionOption):
        value = stream
    else:
        stream = stream.replace(' ', '')
        values = stream.split(',')

        if 'local' in values:
            local = True
        else:
            local = False

        if 'remote' in values:
            remote = True
        else:
            remote = False

        value = ActionOption(local=local, remote=remote)

    return SetOrRemoveValue(action=value)


def parse_spam_value(stream: str) -> SpamValue:
    '''Parses the values for the `Spam` header.

    :param stream: String to parse.

    :return: An `SpamValue` instance.

    :raises ParseError: Raised if there is no true/false value, or valid numbers for the score or threshold.
    '''

    stream = stream.replace(' ', '')
    try:
        found, score, threshold = re.split('[;/]', stream)
    except ValueError:
        raise ParseError('Spam header in unrecognizable format')

    found = found.lower()
    if found in ['true', 'yes']:
        value = True
    elif found in ['false', 'no']:
        value = False
    else:
        raise ParseError('Spam header is not a true or false value')

    try:
        score = float(score)
    except ValueError:
        raise ParseError('Cannot parse Spam header score value')

    try:
        threshold = float(threshold)
    except ValueError:
        raise ParseError('Cannot parse Spam header threshold value')

    return SpamValue(value=value, score=score, threshold=threshold)


def parse_user_value(stream: str) -> UserValue:
    return UserValue(name=stream.strip())


def parse_generic_header_value(stream: str) -> GenericHeaderValue:
    '''Parses any user-defined or currently unknown header values.

    :param stream: String to parse.

    :return: A `GenericHeaderValue` instance.
    '''

    return GenericHeaderValue(value=stream.strip())


def parse_header_value(header: str, value: Union[str, bytes]) -> HeaderValue:
    '''Sends the header value stream to the header value parsing function.

    :param header: Name of the header.
    :param value: String or byte stream of the header value.

    :return: The `HeaderValue` instance from the parsing function.
    '''

    if header in header_value_parsers:
        if isinstance(value, bytes):
            value = value.decode('ascii')
        value = header_value_parsers[header](value)
    else:
        if isinstance(value, bytes):
            value = value.decode('utf8')
        value = parse_generic_header_value(value)

    return value


def parse_header(stream: bytes) -> Tuple[str, HeaderValue]:
    '''Splits the header line and sends to the header parsing function.

    :param stream: Byte stream of the header line.

    :return: A tuple of the header name and value.
    '''

    header, _, value = stream.partition(b':')
    header = header.decode('ascii').strip()
    value = parse_header_value(header, value)

    return header, value


def parse_body(stream: bytes, content_length: int) -> bytes:
    '''Parses the body of a message.

    :param stream: Byte stream for the body.
    :param content_length: Expected length of the body in bytes.

    :return: Byte stream of the body.

    :raises NotEnoughDataError: If the length is less than the stream.
    :raises TooMuchDataError: If the length is more than the stream.
    '''

    if len(stream) == content_length:
        return stream
    elif len(stream) < content_length:
        raise NotEnoughDataError
    else:
        raise TooMuchDataError


header_value_parsers = {
    'Compress': parse_compress_value,
    'Content-length': parse_content_length_value,
    'DidRemove': parse_set_remove_value,
    'DidSet': parse_set_remove_value,
    'Message-class': parse_message_class_value,
    'Remove': parse_set_remove_value,
    'Set': parse_set_remove_value,
    'Spam': parse_spam_value,
    'User': parse_user_value
}
'''Mapping for header names to their parsing functions.'''


class RequestParser(Parser):
    '''Sub-class of the parser for requests.'''

    def __init__(self):
        super().__init__(
            delimiter=b'\r\n',
            status_parser=parse_request_status,
            header_parser=parse_header,
            body_parser=parse_body
        )


class ResponseParser(Parser):
    '''Sub-class of the parser for responses.'''

    def __init__(self):
        super().__init__(
            delimiter=b'\r\n',
            status_parser=parse_response_status,
            header_parser=parse_header,
            body_parser=parse_body
        )
